/*--------------------------------------------------------------+
| Author: Daniel Porumbel 2018                                  |
|License: Creative Commons Attribution 3.0 Unported License     |
|         http://creativecommons.org/licenses/by/3.0/           |             
+--------------------------------------------------------------*/
#include "spheres.h"
#include<iostream>
#include<fstream>
#include<iomanip>
using namespace std;

/*--------------------------------------------------------------------------------*/
/*------                    Two Hamming distance functions                    ----*/

int deltaDist(int enter, int exit, int* c) 
{
   return ((c[enter]==1) ? -1: 1) //s[enter]:0->1, gets closer if we have
          +                       //c[enter]==s[enter]==1 after the move
          ((c[exit]==0) ?  -1: 1);//s[exit]:  1->0, gets closer if we have
                                  //c[exit]==s[exit]==0 after the move
   //EQUIVALENT TO:
   //1-2*c[enter]  +  -1+2*c[exit]
   //=2*(c[exit]-c[enter])
}
int dist(int* sol1, int*sol2)
{
    int hamming = 0;
    for(int i=0;i<n;i++)
        if(sol1[i]!=sol2[i])
            hamming ++;
    return hamming;
}

/*--------------------------------------------------------------------------------*/
/*------           Functions/vars for the priority queue of spheres           ----*/

int maxPerSphere = 3;
int runsLaunched = 0;
int instertedCenters = 0;

sphere* prioritQueue = NULL;
sphere* head         = NULL;//reading "head" that can move along prioritQueue


void printPrioritQueue(){
    clog<<"=====";
    sphere*iter;
    iter    = prioritQueue;
    while((iter!=NULL)){
        clog<<iter->obj<<":"<<iter->sumDst<<","<<iter->launched<<";";
        iter = iter->nxt;
    }
    clog<<"====="<<endl;
}
void erasePrioritQueue()
{
    if(prioritQueue==NULL)
        return;
    head    = prioritQueue;
    prioritQueue = NULL;
    while(head->nxt!=NULL){
        sphere*iter = head->nxt;
        delete head; 
        head = iter;
    }
    instertedCenters = 0;
}


void dequeueMaxPriorCenter(int *sol)
{
    runsLaunched ++;
    int positionInQueue = 0;
    head = prioritQueue;
    while((head->launched>=maxPerSphere)&&(head->nxt!=NULL)){
        head = head->nxt;
        positionInQueue ++;
    }
    for(int jj=0;jj<(::n);jj++)
        sol[jj] = head->sol[jj] ;
    head->launched++;
    clog<<"    dequeue("<<head->obj<<","<<head->launched<<";posrank="<<positionInQueue<<") ";
    int i=0,j=0;
    while((i<20)&&(j<n)){
        if(sol[j]){
            clog<<j<<" ";
            i++;
        }
        j++;
    }
    clog<<endl;
}
//return 1 if center is better than a new sphere with obj=objNew and sumDst=sumDstNew
int compCenters(sphere* center, int objNew, int sumDstNew){
    if(center->launched>0)                                  //started to launch from a center
            return 1;                                       //continue till reaching maxPerSphere
    if(center->obj > objNew)                                //we maximize
        return 1;
#ifdef FIFO_RANK_SAME_OBJ_CENTERS
    if((center->obj==objNew))
            return 1;
#else
    if( (center->obj==objNew)&& (center->sumDst < sumDstNew))
            return 1;
#endif
    return 0;
}
#define MAX_SPHERES 1000
void enqueueCenter(int* newCenter, int objNew){
    instertedCenters ++;
#ifdef MAX_SPHERES
    if(instertedCenters==MAX_SPHERES){
        cout<<"Stop insert at "<<MAX_SPHERES<<"spheres ";
        cout.flush();
        head= prioritQueue;
        head = head->nxt;
        while(head->nxt!=NULL){
            sphere*iter = head->nxt;
            delete head; 
            head = iter;
        }
        prioritQueue->launched = 0;
        prioritQueue->nxt = NULL;
        instertedCenters = 0;
    }
    if(instertedCenters>MAX_SPHERES){              //no more insertions
        return;
    }
#endif
    clog<<"enqueue with objNew="<<objNew<<",pos="<<instertedCenters<<"| ";
    int i=0,j=0;
    //clog first (max) 25 elements of new center
    while((i<25)&&(j<n)){
        if(newCenter[j]){
            clog<<j<<" ";
            i++;
        }
        j++;
    }
    clog<<endl;
    sphere* newElem;
    int sumDstNew = 0;
    if(prioritQueue==NULL){
        prioritQueue = new sphere;
        head    = prioritQueue;
        newElem = prioritQueue;
        newElem->nxt=NULL;
    }else{
        sphere* iter    = prioritQueue;
#ifndef FIFO_RANK_SAME_OBJ_CENTERS
        //update sumDst for existing spheres calculate sumDstNew of new sphere
        while((iter!=NULL)){
            int newDist  = dist(newCenter,iter->sol);
            sumDstNew     += newDist;
            iter->sumDst+= newDist;
            iter = iter->nxt;
        }
#endif
        iter = prioritQueue;
        while((iter->nxt!=NULL)&&(compCenters(iter->nxt,objNew,sumDstNew))){
                iter = iter->nxt;
                //clog<<"jump to add\n";
        }
        //clog<<"adding with sumDstNew="<<sumDstNew<<"\n";
        newElem   = new sphere;
        newElem->nxt = iter->nxt;
        iter->nxt    = newElem;
        if(prioritQueue->nxt==NULL)
            prioritQueue->nxt = newElem;
    }
    
    newElem->sol=new int[::n];
    for(int ii=0;ii<(::n);ii++)
        newElem->sol[ii] = newCenter[ii];
    newElem->obj = objNew;
    newElem->launched = 0;
    newElem->sumDst = sumDstNew;
}
